# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring,unused-import,reimported
from foran.report import generate_report_list


def test_generate_report_list_nothing_ast_empty():
    assert generate_report_list('foo:', False, [], '*') == []


def test_generate_report_list_ast_empty():
    assert generate_report_list('', True, [], '*') == ['\n']


def test_generate_report_list_default_single():
    assert generate_report_list('foo:', True, ['bar']) == ['foo:\n', ' * bar\n']


def test_generate_report_list_ast_single():
    assert generate_report_list('foo:', True, ['bar'], '*') == ['foo:\n', ' * bar\n']


def test_generate_report_list_plus_single():
    assert generate_report_list('foo:', True, ['bar'], '+') == ['foo:\n', ' + bar\n']


def test_generate_report_list_dash_single():
    assert generate_report_list('foo:', True, ['bar'], '-') == ['foo:\n', ' - bar\n']


def test_generate_report_list_default_two():
    assert generate_report_list('foo:', True, ['bar', 'baz']) == ['foo:\n', ' * bar\n * baz\n']


def test_generate_report_list_nothing_default_two():
    assert generate_report_list('foo:', False, ['bar', 'baz']) == []



# -*- coding: utf-8 -*-
# pylint: disable=expression-not-assigned,line-too-long
"""In front or behind (Foran eller bagved)? rendering interface for templates."""
from enum import Enum, auto


class Template(Enum):
    NONE = auto()
    STRING = auto()
    F_STRING = auto()
    JINJA_STRING = auto()
    JINJA_PATH = auto()



# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring,unused-import,reimported
import pathlib

import pytest
from git import Repo
from git.exc import InvalidGitRepositoryError

import foran.foran as fb
from foran.status import Status


def test_main_empty():
    message = 'received wrong number of arguments'
    with pytest.raises(UserWarning) as ex:
        fb.main([]) == 2
        assert message in str(ex.value)


def test_main_few():
    message = 'received wrong number of arguments'
    with pytest.raises(UserWarning) as ex:
        fb.main(['diff', '.', 'would_be_here']) == 2
        assert message in str(ex.value)


def test_main_unknown_command():
    message = 'received unknown command'
    with pytest.raises(UserWarning) as ex:
        fb.main(['unknown', '.', 'would_be_here', '']) == 2
        assert message in str(ex.value)


def test_main_diff():
    assert fb.main(['diff', '.', 'foran-eller-bagved.txt', '']) == 0


def test_main_label():
    assert fb.main(['label', '.', 'STD_OUT', '']) == 0


def test_main_template():
    message = 'ignoring template: ignored'
    with pytest.raises(UserWarning) as ex:
        fb.main(['diff', '.', 'foran-eller-bagved.txt', 'ignored']) == 0
        assert message in str(ex.value)


def test_local_commits_none():
    repo = Repo('.')
    status = Status(repo)
    status.foran = True
    fb.local_commits(repo, status)
    assert status.local_commits == []


def test_local_commits_some():
    repo = Repo('.')
    status = Status(repo)
    status.foran = False
    fb.local_commits(repo, status)
    assert status.local_commits == []


def test_local_commits_no_repo():
    with pytest.raises(InvalidGitRepositoryError):
        _ = Repo('tests')
        # We cannot currently really probe for a git repo that raises GitCommandError


def test_local_commits_no_remote():
    path_to_non_remote = pathlib.Path('tests', 'fixtures', 'non_remote')
    repo = Repo(path_to_non_remote)
    status = Status(repo)
    status.foran = False
    message = 'No remote found, so all commit differences hypothetical'
    with pytest.raises(UserWarning) as ex:
        fb.local_commits(repo, status)
        assert status.local_commits == [message]
        assert message in str(ex.value)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=line-too-long
"""Commandline API gateway for foran."""
import pathlib
import sys
from typing import List, Union

import typer

import foran
import foran.foran as fb
from foran.render import Template

CWD = '.'
APP_NAME = 'In front or behind (Foran eller bagved)?'
APP_ALIAS = 'foran'
app = typer.Typer(
    add_completion=False,
    context_settings={'help_option_names': ['-h', '--help']},
    no_args_is_help=True,
)


@app.callback(invoke_without_command=True)
def callback(
    version: bool = typer.Option(
        False,
        '-V',
        '--version',
        help='Display the foran version and exit',
        is_eager=True,
    )
) -> None:
    """
    In front or behind (Foran eller bagved)?

    Identify and diff a local repository against the remote using templates for reports.
    """
    if version:
        typer.echo(f'{APP_NAME} version {foran.__version__}')
        raise typer.Exit()


def classify_template(template: str) -> Template:
    """
    Process the received template and return the classification
    """
    if not template:
        return Template.NONE

    if '{{' in str(template):
        return Template.JINJA_STRING

    if '{' in str(template):
        return Template.F_STRING

    if pathlib.Path(str(template)).is_file():
        return Template.JINJA_PATH

    return Template.STRING


@app.command('diff')
def diff_repo(
    source: str = typer.Argument(CWD),
    input: str = typer.Option('', '-i', '--input'),
    output: str = typer.Option('', '-o', '--output'),
    template: str = typer.Option('', '-t', '--template'),
) -> int:
    """
    Diff the local against the remote repository state  (no template handling yet)
    """
    repo_root = input if input else source
    target = 'STD_OUT' if not str(output) else str(output)
    return fb.main(['diff', repo_root, target, template])


@app.command('label')
def label_repo(
    source: str = typer.Argument(CWD),
    input: str = typer.Option('', '-i', '--input'),
    output: str = typer.Option('', '-o', '--output'),
    template: str = typer.Option('', '-t', '--template'),
) -> int:
    """
    Labels the local repository state (no template handling yet)
    """
    repo_root = input if input else source
    target = 'STD_OUT' if not str(output) else str(output)
    return fb.main(['label', repo_root, target, template])


@app.command('template')
def app_template(output: str = typer.Option('', '-o', '--output')) -> None:
    """
    Output an example jinja template representing the defaults
    """
    target = 'STD_OUT' if not output else output
    typer.echo(f'Example template generated per {target}')


@app.command('version')
def app_version() -> None:
    """
    Display the foran version and exit
    """
    callback(True)


# pylint: disable=expression-not-assigned
# @app.command()
def main(argv: Union[List[str], None] = None) -> int:
    """Delegate processing to functional module."""
    argv = sys.argv[1:] if argv is None else argv
    return fb.main(argv)



# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring,unused-import,reimported
import pytest
from typer.testing import CliRunner

import foran
import foran.cli as cli
from foran.cli import app
from foran.render import Template

runner = CliRunner()


def test_app_version():
    result = runner.invoke(app, ['version'])
    assert result.exit_code == 0
    assert cli.APP_NAME in result.stdout
    assert foran.__version__ in result.stdout


def test_app_template():
    result = runner.invoke(app, ['template'])
    assert result.exit_code == 0
    assert 'Example template generated per ' in result.stdout


def test_cli_main():
    message = 'ignoring template: ignored'
    with pytest.raises(UserWarning) as ex:
        cli.main(['diff', '.', 'foran-eller-bagved.txt', 'ignored']) == 0
        assert message in str(ex.value)


def test_cli_diff_path():
    message = 'ignoring template: ignored'
    with pytest.raises(UserWarning) as ex:
        cli.diff_repo(source=cli.CWD, output='foran-eller-bagved.txt') == 0
        assert message in str(ex.value)


def test_cli_diff_stdout():
    message = 'ignoring template: ignored'
    with pytest.raises(UserWarning) as ex:
        cli.diff_repo(source=cli.CWD) == 0
        assert message in str(ex.value)


def test_cli_label_path():
    message = 'ignoring template: ignored'
    with pytest.raises(UserWarning) as ex:
        cli.label_repo(source=cli.CWD, output='foran-eller-bagved.txt') == 0
        assert message in str(ex.value)


def test_cli_label_stdout():
    message = 'ignoring template: ignored'
    with pytest.raises(UserWarning) as ex:
        cli.label_repo(source=cli.CWD) == 0
        assert message in str(ex.value)


def test_cli_classify_template_empty():
    cli.classify_template('') == Template.NONE


def test_cli_classify_template_jinja_string():
    cli.classify_template('{{ JINJA }}') == Template.JINJA_STRING


def test_cli_classify_template_f_string():
    cli.classify_template('{f_string}') == Template.F_STRING


def test_cli_classify_template_jinja_path():
    cli.classify_template('README.md') == Template.JINJA_PATH


def test_cli_classify_template_string():
    cli.classify_template('NOTHING_ELSE_MATTERS') == Template.STRING



# -*- coding: utf-8 -*-
# pylint: disable=expression-not-assigned,line-too-long
"""In front or behind (Foran eller bagved)? API."""
import os
import pathlib
import warnings
from typing import List, Union

from git import Repo
from git.exc import GitCommandError

from foran.report import Format, Report, report_as
from foran.status import Status

DEBUG_VAR = 'FORAN_DEBUG'
DEBUG = os.getenv(DEBUG_VAR)

ENCODING = 'utf-8'
ENCODING_ERRORS_POLICY = 'ignore'


def local_commits(repo: Repo, status: Status) -> None:
    """Yes"""
    if not status.foran:
        branch = repo.active_branch.name
        try:
            status.local_commits = [rev for rev in repo.iter_commits(f'origin/{branch}..{branch}')]
        except GitCommandError as ex:
            symptom = ex.stderr.replace('\n', '')
            if "fatal: bad revision 'origin/default..default'" in symptom:
                warning = 'No remote found, so all commit differences hypothetical'
                status.local_commits = [warning]
                warnings.warn(warning)
            else:
                raise


def local_staged(repo: Repo, status: Status) -> None:
    """Truly"""
    status.local_staged = [item.a_path for item in repo.index.diff('HEAD')]


def local_files(repo: Repo, status: Status) -> None:
    """Sure"""
    status.local_files = [item.a_path for item in repo.index.diff(None)]


def main(argv: Union[List[str], None] = None) -> int:
    """Drive the analysis."""
    # [('diff'|'label'), repo_root, target, template]
    if not argv or len(argv) != 4:
        warnings.warn('received wrong number of arguments')
        return 2

    if argv[0] not in ('diff', 'label'):
        warnings.warn('received unknown command')
        return 2

    command, repo_root, target, template = argv

    if not pathlib.Path(str(repo_root)).is_dir():
        warnings.warn('repository root is no directory')
        return 1

    if template and target != 'STD_OUT':
        warnings.warn('templates not yet implemented')

    report = Report(stem=target, file_format=Format.NONE)

    repo = Repo(repo_root)
    status = Status(repo)
    if command == 'diff':
        local_commits(repo, status)
        local_staged(repo, status)
        local_files(repo, status)
    report_as(status, report)
    return 0
